(function() {
    "use strict";

    // Glossary
    //
    // Stores a list of terms and definitions.  Includes routines for finding terms in strings, and
    // returning formatted content.  This class is used by both the runtime engine and the standalone
    // glossary page
    //
    // note the use of an array to store the glossary items.  this is necessary because we expect the
    // glossary file itself to "rank" the order in which terms should be searched for--it is often the
    // case that the glossary will contain both a base item and an extended item; for example, the
    // terms "foo" and "foo bar".  authors should place terms that include subterms earlier in the
    // xml document (that is, define "foo bar" before defining "foo") or in the code that uses the
    // add() method.
    //
    //
    //



    function Glossary() {

        // collections
        this._items = [];
        this._exclusions = {};

        // simple vars
        this.ignoreLinkText = true;
        this.matchAllCapsExactly = true;
        this._regexTemplateIgnoreLinks = "(?!<a[^>]*?>)(?:\\b)%TERM%(?!\\w)(?![^<]*?<\/a>)";
        this._regexTemplate = "(?:\\b)%TERM%(?!\\w)";

    }


    Glossary.prototype.load = function(filename, loadAsync) {

        var xhr = new XMLHttpRequest();
        var instance = this;

        loadAsync = !!loadAsync; // force to boolean, default is false

        try {
            xhr.open("Get", filename, loadAsync);
            xhr.onload = function() {
                if (xhr.readyState === 4) {
                    if (xhr.status === 200) {
                        instance._loadXml(xhr.responseXML);
                    } else {
                        instance.onloaderror(xhr.statusText);
                        instance.onloadcomplete();
                    }
                }
            };
            xhr.onerror = function(exception) {
                instance.onloaderror(exception.message);
                instance.onloadcomplete();
            };
            xhr.send();
        } catch (ex) {
            instance.onloaderror(ex.message);
            instance.onloadcomplete();
        }

    };
    Glossary.prototype._loadXml = function(xml) {

        if (xml && xml.firstChild) {
            var xmlItems = xml.firstChild.getElementsByTagName("item");
            for (var i = 0; i < xmlItems.length; i++) {
                var term = xmlItems[i].attributes.getNamedItem("term").value;
                var definition = xmlItems[i].textContent;
                this.add(term, definition);
            }
        }

        this.onloadsuccess();
        this.onloadcomplete();
    };
    Glossary.prototype.onloadsuccess = function() {
        if (window["$"]) window["$"](this).trigger(Glossary.LOAD_SUCCESS);
    };
    Glossary.prototype.onloaderror = function(statusText) {
        if (window["$"]) {
            var event = window["$"].Event(Glossary.LOAD_ERROR);
            event["message"] = statusText;
            window["$"](target).trigger(event);
        }
    };
    Glossary.prototype.onloadcomplete = function() {
        if (window["$"]) window["$"](this).trigger(Glossary.LOAD_COMPLETE);
    };


    Glossary.prototype.add = function(term, definition) {
    // adds (or replaces) the term in our array of items

        if (isNullOrEmpty(term) || (isNullOrEmpty(definition))) return;

        var item = this._newItem(term, definition);
        var index = indexOfTerm(term, this._items);
        if (index < 0) {
            this._items.push(item);
        } else {
            this._items[index] = item;
        }


    };
    Glossary.prototype._newItem = function(term, definition) {
    // returns a new glossary item object.  no error checking!
        return {
            term: term,
            definition: definition,
            key: term.toLowerCase()
        };
    };
    Glossary.prototype.contains = function(term) {
        return (indexOfTerm(term, this._items) >= 0);
    };
    Glossary.prototype.definitionOf = function(term) {
        var index = indexOfTerm(term, this._items);
        return (index < 0)
            ? ""
            : this._items[index].definition;
    };
    Glossary.prototype.clear = function() {
        this._items = [];
        this._exclusions = {};
    };
    Glossary.prototype.excludeTerm = function(term) {
        if (!isNullOrEmpty(term)) this._exclusions[term] = [term];
    };
    Glossary.prototype.isExcluded = function(term) {
        return (this._exclusions[term] !== undefined);
    };
    Glossary.prototype.clearExclusions = function() {
        this._exclusions = {};
    };

    Glossary.prototype._getRegExpStringOf = function(term) {
        // returns regexp string to use for the passed term

        var template = (this.ignoreLinkText)
            ? this._regexTemplateIgnoreLinks
            : this._regexTemplate;

        term = escapeTerm(term);
        return stringReplaceAll(template, Glossary.TERM_TOKEN, term);
    };
    Glossary.prototype._getRegExpFlagsOf = function(term, firstOccurrenceOnly) {
        // returns the appropriate flags to use

        firstOccurrenceOnly = !!firstOccurrenceOnly; // force to boolean

        var flags = (firstOccurrenceOnly)
            ? ""
            : "g";

        if ((!this.matchAllCapsExactly) || (term != term.toUpperCase())) {
            flags += "i";
        }

        return flags;
    };
    Glossary.prototype._getRegExpOf = function(term, firstOccurrenceOnly) {
        // returns a RegExp object for the passed term
        return new RegExp(this._getRegExpStringOf(term), this._getRegExpFlagsOf(term, firstOccurrenceOnly));
    };

    Glossary.prototype.termIn = function(text, term) {
        // simple check to see if a term exists in the text
        return (isNullOrEmpty(text) || isNullOrEmpty(term))
            ? false
            : this._getRegExpOf(term, true).test(text);
    };
    Glossary.prototype.termsIn = function(text, ignoreExclusions) {
        // returns an array of the terms found in text.  to ensure that we don't mistakenly identify
        // items twice (think finding "foo" and then "foo bar"), we'll remove each item found after
        // we test for it

        if (isNullOrEmpty(text)) return [];
        ignoreExclusions = !!ignoreExclusions; // force to boolean

        var matches = [];
        for (var i = 0; i < this._items.length; i++) {
            var term = this._items[i].term;
            if (ignoreExclusions || (this.isExcluded(term) == false)) {
                var re = this._getRegExpOf(term, false);
                if (re.test(text)) {
                    matches.push(term);
                    text = text.replace(re,"");
                }
            }
        }

        matches.sort(caseInsensitiveStringCompare);
        return matches;
    };
    Glossary.prototype.termsInPattern = function(pattern, ignoreExclusions) {
        // returns an array of the terms matching pattern, which is either
        // a RegExp string or a RegExp object

        // simple cases
        if (isNullOrEmpty(pattern)) return [];
        ignoreExclusions = !!ignoreExclusions; // force to boolean

        if (typeof pattern == "string") {
            return this.termsInPattern(new RegExp(pattern, "i"), ignoreExclusions);

        } else if (pattern instanceof RegExp) {

            var matches = [];
            for (var i = 0; i < this._items.length; i++) {
                var term = this._items[i].term;
                if (ignoreExclusions || (this.isExcluded(term) == false)) {
                    if (pattern.test(term)) matches.push(term);
                }
            }

            matches.sort(caseInsensitiveStringCompare);
            return matches;

        } else {
            Glossary.trace("unsupported pattern object passed to termsInPattern()");
        }

    };

    Glossary.prototype.formatTerm = function(term, template) {
    // returns a string of the term formatted by the template

        if (isNullOrEmpty(template)) template = Glossary.DEFAULT_FORMAT_TEMPLATE;

        var definition = this.definitionOf(term);
        if (definition == "") definition = "[no definition found]";

        var result = template;
        result = stringReplaceAll(result, Glossary.TERM_TOKEN, term);
        result = stringReplaceAll(result, Glossary.DEFINITION_TOKEN, definition);

        return result;
    };
    Glossary.prototype.formatTerms = function(terms, template) {
        // returns a string concatenated across the array of terms

        var result = "";

        for (var i = 0; i < terms.length; i++) {
            result += this.formatTerm(terms[i], template);
        }

        return result;
    };
    Glossary.prototype.replaceTermsIn = function(text, template, firstOccurrenceOnly, ignoreExclusions) {
        // returns a string where the terms found in text are replaced by the tokenized template string
        //
        // to ensure  that we don't mistakenly identify items twice (think finding "foo" and then "foo bar"),
        // we'll initially replace each term with a GUID.  Once all terms have been replaced, we'll make
        // another sweep through the text, replacing the GUIDs with the actual values

        // parameter checks, cleanup
        if (isNullOrEmpty(text) || isNullOrEmpty(template)) return "";
        firstOccurrenceOnly = !!firstOccurrenceOnly; // force to boolean
        ignoreExclusions = !!ignoreExclusions; // force to boolean

        // iterate through the terms, replacing with GUIDs as appropriate
        var placeholders = [];
        var i;

        for (i = 0; i < this._items.length; i++) {
            var term = this._items[i].term;
            var definition = this._items[i].definition;
            if (ignoreExclusions || (this.isExcluded(term) == false)) {
                // process this term
                //
                // since .exec() only returns the first match, we need to iterate over the string if
                // we're replacing all occurrences.  this is why we build each RegExp object without
                // the global flag.
                //
                // we iterate until the .exec() call returns null.
                //
                var result = null;
                do {
                    var re = this._getRegExpOf(term, true); // .exec() only does the first occurrence anyway
                    result = re.exec(text);
                    if (result) {
                        var placeholder = getReplacementPlaceholder(result, term, definition, template);
                        placeholders.push(placeholder);
                        text = text.replace(re, placeholder.source);
                    }
                    if (firstOccurrenceOnly) result = null; // force early termination
                } while (result != null);
            }
        }

        // now that we've found all the terms and replaced them with
        // placeholders, we'll loop through the placeholders and replace
        // them with the original results
        //
        for (i = 0; i < placeholders.length; i++) {
            text = stringReplaceAll(text, placeholders[i].source, placeholders[i].replace);
        }


        return text;
    };

    Glossary.LOAD_SUCCESS = "glossaryLoadSuccess";
    Glossary.LOAD_ERROR = "glossaryLoadError";
    Glossary.LOAD_COMPLETE = "glossaryLoadComplete";
    Glossary.TERM_TOKEN = "%TERM%";
    Glossary.DEFINITION_TOKEN = "%DEFINITION%";
    Glossary.SOURCE_TOKEN = "%SOURCE%";
    Glossary.DEFAULT_FORMAT_TEMPLATE = '<div class="term">%TERM%</div><div class="definition">%DEFINITION%</div>';
    Glossary.DEFAULT_REPLACE_TEMPLATE = '<a href="#" class="glossary-term-link" onclick="showGlossaryTerm(\'%TERM%\')">%SOURCE%</a>';

    // utility methods
    Glossary.trace = function(message){try{console&&trace&&trace(message)}catch(ex){}};
    Glossary.getQuerystringParameter = function(key, defaultValue) {

        if (window.location.search.length < 1) return defaultValue;

        // get querystring pairs into an array
        var queryString = window.location.search.substring(1);
        var pairs = queryString.split("&");

        // iterate through the array, looking for the key
        for (var i = 0; i < pairs.length; i++) {
            var pair = pairs[i].split("=");
            if ((pair.length > 1) && (pair[0].toLowerCase() == key.toLowerCase())) {
                return decodeURIComponent(pair[1]);
            }
        }

        // if we're here, we didn't find it
        return defaultValue;
    };


    window["Glossary"] = Glossary;


//region.  glossary-specific support (closure) routines

    function indexOfTerm(term, items) {
        // returns the index of the item containing term in items, or -1

        if (isNullOrEmpty(term) || !items || items.length == 0) return -1;

        for (var i = 0; i < items.length; i++) {
            if (items[i].term == term) return i;
        }

        return -1; // not found
    }

    function escapeTerm(term) {
        // returns a version of term that has regexp control characters escaped

        term = stringReplaceAll(term, "\\", "\\\\");
        term = stringReplaceAll(term, "\\\\r", "\\r");
        term = stringReplaceAll(term, "\\\\n", "\\n");
        term = stringReplaceAll(term, "\\\\t", "\\t");
        term = stringReplaceAll(term, "\\\\f", "\\f");

        term = stringReplaceAll(term, ".", "\\.");
        term = stringReplaceAll(term, "+", "\\+");
        term = stringReplaceAll(term, "*", "\\*");
        term = stringReplaceAll(term, "?", "\\?");
        term = stringReplaceAll(term, "^", "\\^");
        term = stringReplaceAll(term, "$", "\\$");

        term = stringReplaceAll(term, "(", "\\(");
        term = stringReplaceAll(term, ")", "\\)");
        term = stringReplaceAll(term, "[", "\\[");
        term = stringReplaceAll(term, "]", "\\]");
        term = stringReplaceAll(term, "{", "\\{");
        term = stringReplaceAll(term, "}", "\\}");

        term = stringReplaceAll(term, "|", "\\|");
        term = stringReplaceAll(term, "/", "\\/");
        term = stringReplaceAll(term, "'", "\\'");
        term = stringReplaceAll(term, "#", "\\#");

        return term;

    }

    function caseInsensitiveStringCompare(a, b) {
        // comparison function for two strings.  like all comparison functions, we return
        // -1 if a is less than b ("bar"<"foo")
        //  1 if a is greater than b ("foo">"bar")
        //  0 if they are equal

        a = a.toLowerCase();
        b = b.toLowerCase();

        if (a < b) {
            return -1;
        } else if (a > b) {
            return 1;
        } else {
            return 0;
        }
    }

    /*
    function caseInsensitiveItemCompare(a, b) {
        // where a and b are glossary items
        return caseInsensitiveStringCompare(a.term, b.term);
    }
    */

    function getReplacementPlaceholder(execResult, term, definition, template) {
        // generates a replacement placeholder object based on the
        // results from a RegExp.exec() call.  NO PARAMETER VALIDATION!

        var replaceText = template;
        replaceText = stringReplaceAll(replaceText, Glossary.SOURCE_TOKEN, execResult[0]);
        replaceText = stringReplaceAll(replaceText, Glossary.TERM_TOKEN, term);
        replaceText = stringReplaceAll(replaceText, Glossary.DEFINITION_TOKEN, definition);

        return {
            source: newGuid("D"),
            replace: replaceText
        }


    }
//endregion

//region.  generic support routines

    // singleline generics
    function stringReplaceAll(source, item, replace){ return source.split(item).join(replace);	}
    function isNullOrEmpty(source){return null==source||""==source}


    // more complex generics
    function newGuid(format) {
        // rfc 4122 version 4 conformant

        if (isNullOrEmpty(format)) format = "D";

        // support methods
        //
        var randomSpecialChar = function () {
            var specialChars = ["8", "9", "A", "B"];
            return specialChars[Math.floor(Math.random() * 4)];
        };

        var randomHexDigitString = function (length) {

            var result = "";
            for (var i = 0; i < length; i++) {
                result += Math.floor(Math.random() * 16).toString(16);
            }

            return result;
        };

        var guid = randomHexDigitString(8) + "-"
            + randomHexDigitString(4) + "-"
            + "4" + randomHexDigitString(3) + "-"
            + randomSpecialChar() + randomHexDigitString(4) + "-"
            + randomHexDigitString(12);


        switch (format) {
            case "D":
                return guid.toUpperCase();
            case "d":
                return guid.toLowerCase();
            case "N":
                guid = stringReplaceAll(guid, "-", "");
                return guid.toUpperCase();
            case "n":
                guid = stringReplaceAll(guid, "-", "");
                return guid.toLowerCase();
            case "B":
                guid = "{" + guid + "}";
                return guid.toUpperCase();
            case "b":
                guid = "{" + guid + "}";
                return guid.toLowerCase();
            case "P":
                guid = "(" + guid + ")";
                return guid.toUpperCase();
            case "p":
                guid = "(" + guid + ")";
                return guid.toLowerCase();

            default:
                return guid.toUpperCase();
        }

    }



//endregion

}());
//# sourceURL=Glossary.js
